package xinlake.platform;

import android.content.ContentUris;
import android.content.Context;
import android.content.res.AssetManager;
import android.database.Cursor;
import android.net.Uri;
import android.os.Environment;
import android.provider.DocumentsContract;
import android.provider.MediaStore;
import android.provider.OpenableColumns;
import android.webkit.MimeTypeMap;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;


/**
 * TODO test
 * @author Xinlake Liu, Opensource
 * @version 2021.10
 */
public final class XinFile {
    private static final String Tag = "XinFile";

    /**
     * No additional permissions are required for the calling app to read or write files under the returned path.
     * @param assetFile The name of the asset to open. This name can be hierarchical.
     * @return Return a path to file which may be read in common way.
     */
    public static @Nullable
    String cacheAssetFile(@NonNull Context context, @NonNull String assetFile, boolean overwrite) {
        AssetManager assetManager = context.getAssets();
        try {
            File cacheFile = new File(context.getCacheDir(), assetFile);
            if (cacheFile.exists() && cacheFile.length() > 0 && !overwrite) {
                return cacheFile.getAbsolutePath();
            }

            // Read data from assets.
            InputStream inputStream = assetManager.open(assetFile);
            BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);

            byte[] data = new byte[bufferedInputStream.available()];
            int read = bufferedInputStream.read(data);
            bufferedInputStream.close();
            inputStream.close();
            if (read < 1) {
                return null;
            }

            // Create copy file in storage.
            FileOutputStream fileOutputStream = new FileOutputStream(cacheFile);
            fileOutputStream.write(data);
            fileOutputStream.close();

            // Return a path to file which may be read in common way.
            return cacheFile.getAbsolutePath();
        } catch (Exception exception) {
            exception.printStackTrace();
        }

        return null;
    }

    public static void cacheFromUri(Context context, Uri uri, String destinationPath) {
        InputStream inputStream = null;
        BufferedOutputStream bufferedOutputStream = null;
        try {
            inputStream = context.getContentResolver().openInputStream(uri);
            bufferedOutputStream = new BufferedOutputStream(
                new FileOutputStream(destinationPath, false));

            byte[] buf = new byte[1024];
            while (true) {
                int length = inputStream.read(buf);
                if (length < 0) {
                    break;
                }
                bufferedOutputStream.write(buf, 0, length);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

        try {
            if (inputStream != null) inputStream.close();
            if (bufferedOutputStream != null) bufferedOutputStream.close();
        } catch (IOException exception) {
            exception.printStackTrace();
        }
    }

    /**
     * @param textFile UTF-8 string lines
     * @return the string lines split by LF
     */
    public static @Nullable
    String[] readAssetText(@NonNull Context context, @NonNull String textFile) {
        AssetManager assetManager = context.getAssets();
        try {
            // Read data from assets.
            InputStream inputStream = assetManager.open(textFile);
            BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);

            byte[] data = new byte[bufferedInputStream.available()];
            int read = bufferedInputStream.read(data);
            bufferedInputStream.close();
            inputStream.close();
            if (read < 1) {
                return null;
            }

            String lines = new String(data, StandardCharsets.UTF_8);
            return lines.split("\n");
        } catch (Exception exception) {
            exception.printStackTrace();
        }

        return null;
    }

    public static long calculateFolderSize(File directory) {
        long length = 0;
        File[] files = directory.listFiles();
        if (files != null) {
            for (File file : files) {
                if (file.isFile()) {
                    length += file.length();
                } else {
                    length += calculateFolderSize(file);
                }
            }
        }
        return length;
    }

    public static int calculateFileCount(File file) {
        return calculateFileCount(file, 0);
    }

    /**
     * Recursively count all files in the directory include it's subtree.
     * @param countSoFar file count of previous counting
     * @param directory  The root of the tree to count.
     */
    private static int calculateFileCount(File directory, int countSoFar) {
        if (!directory.isDirectory()) {
            countSoFar++;
            return countSoFar;
        }

        final String[] fileList = directory.list();
        if (fileList == null) {
            return countSoFar;
        }
        for (String fileName : fileList) {
            File file = new File(directory.getAbsolutePath() + File.separator + fileName);
            countSoFar += calculateFileCount(file, 0);
        }
        return countSoFar;
    }

    public static String getName(String filename) {
        if (filename == null) {
            return null;
        }
        int index = filename.lastIndexOf('/');
        if (index < 0) {
            return filename;
        }
        return filename.substring(index + 1);
    }

    /**
     * Gets the extension of a file name, like ".png" or ".jpg".
     * @return Extension including the dot("."); "" if there is no extension;
     * null if uri was null.
     */
    public static String getExtension(String path) {
        if (path == null) {
            return null;
        }

        int dot = path.lastIndexOf(".");
        if (dot >= 0) {
            return path.substring(dot);
        } else {
            // No extension.
            return "";
        }
    }

    /**
     * @return The MIME type for the given file.
     */
    public static String getMimeType(File file) {
        String extension = getExtension(file.getName());
        if (extension.length() > 0) {
            return MimeTypeMap.getSingleton().getMimeTypeFromExtension(extension.substring(1));
        }

        return "application/octet-stream";
    }

    public static String getFileName(@NonNull Context context, Uri uri) {
        String mimeType = context.getContentResolver().getType(uri);
        String filename = null;

        if (mimeType == null) {
            String path = getPath(context, uri);
            if (path == null) {
                filename = getName(uri.toString());
            } else {
                File file = new File(path);
                filename = file.getName();
            }
        } else {
            Cursor returnCursor = context.getContentResolver().query(uri, null,
                null, null, null);
            if (returnCursor != null) {
                int nameIndex = returnCursor.getColumnIndex(OpenableColumns.DISPLAY_NAME);
                returnCursor.moveToFirst();
                filename = returnCursor.getString(nameIndex);
                returnCursor.close();
            }
        }

        return filename;
    }

    /**
     * @return The MIME type for the give Uri.
     */
    public static String getMimeType(Context context, Uri uri) {
        File file = new File(getPath(context, uri));
        return getMimeType(file);
    }

    /**
     * @return The MIME type for the give String Uri.
     */
    public static String getMimeType(Context context, String url) {
        String type = context.getContentResolver().getType(Uri.parse(url));
        if (type == null) {
            type = "application/octet-stream";
        }
        return type;
    }

    /**
     * Get the value of the data column for this Uri. This is useful for
     * MediaStore Uris, and other file-based ContentProviders.
     * @param context       The context.
     * @param uri           The Uri to query.
     * @param selection     (Optional) Filter used in the query.
     * @param selectionArgs (Optional) Selection arguments used in the query.
     * @return The value of the _data column, which is typically a file path.
     */
    public static String getDataColumn(Context context, Uri uri, String selection,
                                       String[] selectionArgs) {
        Cursor cursor = null;
        final String column = MediaStore.Files.FileColumns.DATA;
        final String[] projection = {
            column
        };

        try {
            cursor = context.getContentResolver().query(uri, projection, selection, selectionArgs,
                null);
            if (cursor != null && cursor.moveToFirst()) {
                final int column_index = cursor.getColumnIndexOrThrow(column);
                return cursor.getString(column_index);
            }
        } catch (Exception exception) {
            exception.printStackTrace();
        } finally {
            if (cursor != null) {
                cursor.close();
            }
        }
        return null;
    }

    /**
     * Get a file path from a Uri. This will get the the path for Storage Access
     * Framework Documents, as well as the _data field for the MediaStore and
     * other file-based ContentProviders.<br>
     * <br>
     * Callers should check whether the path is local before assuming it
     * represents a local file.
     * @param context The context.
     * @param uri     The Uri to query.
     */
    public static String getPath(final Context context, final Uri uri) {
        String absolutePath = getLocalPath(context, uri);
        return absolutePath != null ? absolutePath : uri.toString();
    }

    private static String getLocalPath(final Context context, final Uri uri) {
        // DocumentProvider
        if (DocumentsContract.isDocumentUri(context, uri)) {
            // ExternalStorageProvider
            if ("com.android.externalstorage.documents".equals(uri.getAuthority())) {
                final String docId = DocumentsContract.getDocumentId(uri);
                final String[] split = docId.split(":");
                final String type = split[0];

                if ("primary".equalsIgnoreCase(type)) {
                    return Environment.getExternalStorageDirectory() + "/" + split[1];
                } else if ("home".equalsIgnoreCase(type)) {
                    return Environment.getExternalStorageDirectory() + "/documents/" + split[1];
                }
            }
            // DownloadsProvider
            else if ("com.android.providers.downloads.documents".equals(uri.getAuthority())) {
                final String id = DocumentsContract.getDocumentId(uri);
                if (id != null && id.startsWith("raw:")) {
                    return id.substring(4);
                }

                String[] contentUriPrefixesToTry = new String[]{
                    "content://downloads/public_downloads",
                    "content://downloads/my_downloads"
                };
                for (String contentUriPrefix : contentUriPrefixesToTry) {
                    try {
                        Uri contentUri = ContentUris.withAppendedId(Uri.parse(contentUriPrefix), Long.parseLong(id));
                        String path = getDataColumn(context, contentUri, null, null);
                        if (path != null) {
                            return path;
                        }
                    } catch (Exception exception) {
                        exception.printStackTrace();
                    }
                }

                // path could not be retrieved using ContentResolver, therefore copy file to accessible cache using streams
                String fileName = getFileName(context, uri);
                File cacheDir = context.getCacheDir();
                File file = generateFileName(fileName, cacheDir);
                String destinationPath = null;
                if (file != null) {
                    destinationPath = file.getAbsolutePath();
                    cacheFromUri(context, uri, destinationPath);
                }

                return destinationPath;
            }
            // MediaProvider
            else if ("com.android.providers.media.documents".equals(uri.getAuthority())) {
                final String docId = DocumentsContract.getDocumentId(uri);
                final String[] split = docId.split(":");
                final String type = split[0];

                Uri contentUri = null;
                if ("image".equals(type)) {
                    contentUri = MediaStore.Images.Media.EXTERNAL_CONTENT_URI;
                } else if ("video".equals(type)) {
                    contentUri = MediaStore.Video.Media.EXTERNAL_CONTENT_URI;
                } else if ("audio".equals(type)) {
                    contentUri = MediaStore.Audio.Media.EXTERNAL_CONTENT_URI;
                }

                final String selection = "_id=?";
                final String[] selectionArgs = new String[]{
                    split[1]
                };

                return getDataColumn(context, contentUri, selection, selectionArgs);
            }
        }
        // MediaStore (and general)
        else if ("content".equalsIgnoreCase(uri.getScheme())) {
            return getDataColumn(context, uri, null, null);
        }
        // File
        else if ("file".equalsIgnoreCase(uri.getScheme())) {
            return uri.getPath();
        }

        return null;
    }

    @Nullable
    public static File generateFileName(@Nullable String name, File directory) {
        if (name == null) {
            return null;
        }

        File file = new File(directory, name);
        if (file.exists()) {
            String fileName = name;
            String extension = "";
            int dotIndex = name.lastIndexOf('.');
            if (dotIndex > 0) {
                fileName = name.substring(0, dotIndex);
                extension = name.substring(dotIndex);
            }

            int index = 0;
            while (file.exists()) {
                index++;
                name = fileName + '-' + index + extension;
                file = new File(directory, name);
            }
        }

        return file;
    }
}
